package org.hardknockjava.lesson20.example01.reader.broadway;


/******************************************************************
 * This example uses the Document Object Model to read an XML
 * file, dump the values obtained from it, and reproduce it
 * on the console.
 * 
 * Acknowledgments to Lokesh Gupta and Yong Mook Kim for code examples.
 */

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.OutputStream;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

public class BroadwayReaderDOM {
    private List<ShowPOJO> showList = new ArrayList<ShowPOJO>();

    private DocumentBuilderFactory factory = null;
    private DocumentBuilder builder = null;

    public static void main(String[] args) throws Exception {
        BroadwayReaderDOM instance = new BroadwayReaderDOM();

        // Read the file and store it in instances of Production.
        instance.read("data/broadway.xml");

        // Dump the instances to the console.
        instance.dump();

        // Write the instances to the console.
        System.out.println();
        instance.write();
    }

    public BroadwayReaderDOM() throws Exception {
        factory = DocumentBuilderFactory.newInstance();
        builder = factory.newDocumentBuilder();
    }

    // Read the XML and store it.
    private void read(String fileName) throws Exception {
        File showFile = new File(fileName);
        Document document = builder.parse(showFile);

        // nodeList contains all the <show> nodes.
        NodeList nodeList = document.getElementsByTagName("show");

        // Process each <show> in turn.
        for (int index = 0; index < nodeList.getLength(); index++) {
            Node node = nodeList.item(index);
            if (node.getNodeType() == Node.ELEMENT_NODE) {
                Element showElement = (Element) node;

                // The title is an Attribute of the <show> node.
                String title = showElement.getAttribute(ShowPOJO.FIELD_TITLE);

                // The remaining fields are elements nested within the <show> node.
                List<String> playwrightList = getListValue(ShowPOJO.FIELD_PLAYWRIGHTS, ShowPOJO.FIELD_PLAYWRIGHT,
                        showElement);
                String theatre = getNodeStringValue(ShowPOJO.FIELD_THEATRE, showElement);
                List<String> composerList = getListValue(ShowPOJO.FIELD_COMPOSERS, ShowPOJO.FIELD_COMPOSER,
                        showElement);
                List<String> lyricistList = getListValue(ShowPOJO.FIELD_LYRICISTS, ShowPOJO.FIELD_LYRICIST,
                        showElement);
                Date opening = getNodeDateValue(ShowPOJO.FIELD_OPENING, showElement);
                Date closing = getNodeDateValue(ShowPOJO.FIELD_CLOSING, showElement);
                Integer previews = getNodeIntegerValue(ShowPOJO.FIELD_PREVIEWS, showElement);
                Integer performances = getNodeIntegerValue(ShowPOJO.FIELD_PERFORMANCES, showElement);

                ShowPOJO show = new ShowPOJO(title, theatre, playwrightList, composerList, lyricistList, opening,
                        closing, previews, performances);
                showList.add(show);
            }
        }
    }

    // This method returns the text contained within an Element.
    // Note that there's no way to get a single element by name.
    // The DOM assumes there might be more than one. But we're
    // interested in only the first even if there are more.
    private String getNodeStringValue(String elementName, Element element) {
        String returnValue = null;
        NodeList list = element.getElementsByTagName(elementName);
        if (list.getLength() > 0) {
            Node node = list.item(0);
            returnValue = node.getTextContent();
        }

        return returnValue;
    }

    private List<String> getListValue(String parentName, String childName, Element element) {
        List<String> returnValue = new ArrayList<String>();

        NodeList parentList = element.getElementsByTagName(parentName);

        for (int parentIndex = 0; parentIndex < parentList.getLength(); parentIndex++) {
            Node parentNode = parentList.item(parentIndex);
            NodeList childList = parentNode.getChildNodes();
            for (int childIndex = 0; childIndex < childList.getLength(); childIndex++) {
                Node childNode = childList.item(childIndex);
                if (childNode.getChildNodes().getLength() > 0) {
                    returnValue.add(childNode.getTextContent());
                }
            }
        }

        return returnValue;
    }

    // This method is for elements that contain the string representation
    // of dates.
    private Date getNodeDateValue(String elementName, Element element) {
        String dateStr = getNodeStringValue(elementName, element);
        Date returnValue = null;

        if (dateStr != null) {
            SimpleDateFormat fmt = new SimpleDateFormat("MM/dd/yyyy");
            try {
                returnValue = fmt.parse(dateStr);
            } catch (Exception e) {
                returnValue = null;
            }
        }

        return returnValue;
    }

    // This method is for elements that contain the string representation
    // of Integers.
    private Integer getNodeIntegerValue(String elementName, Element element) {
        String integerStr = getNodeStringValue(elementName, element);
        Integer returnValue = null;

        if (integerStr != null) {
            try {
                returnValue = Integer.valueOf(integerStr);
            } catch (Exception e) {
                returnValue = null;
            }
        }

        return returnValue;
    }

    // Dump the list of Shows to the console.
    private void dump() {
        for (ShowPOJO show : showList) {
            System.out.println(show.toString());
        }
    }

    // Write a new XML file containing the list of Shows,
    // in reverse order.
    private void write() throws Exception {
        // root elements
        Document doc = builder.newDocument();
        Element rootElement = doc.createElement("broadway");
        doc.appendChild(rootElement);

        for (int index = showList.size() - 1; index >= 0; index--) {
            ShowPOJO show = showList.get(index);
            Element showElement = doc.createElement("show");
            showElement.setAttribute(ShowPOJO.FIELD_TITLE, show.getTitle());
            rootElement.appendChild(showElement);

            createChild(ShowPOJO.FIELD_THEATRE, show.getTheatre(), doc, showElement);
            createChild(ShowPOJO.FIELD_PLAYWRIGHTS, ShowPOJO.FIELD_PLAYWRIGHT, show.getPlaywrightList(), doc,
                    showElement);
            createChild(ShowPOJO.FIELD_COMPOSERS, ShowPOJO.FIELD_COMPOSER, show.getComposerList(), doc,
                    showElement);
            createChild(ShowPOJO.FIELD_LYRICISTS, ShowPOJO.FIELD_LYRICIST, show.getLyricistList(), doc,
                    showElement);
            createChild(ShowPOJO.FIELD_OPENING, show.getOpening(), doc, showElement);
            createChild(ShowPOJO.FIELD_CLOSING, show.getClosing(), doc, showElement);
            createChild(ShowPOJO.FIELD_PREVIEWS, show.getPreviews(), doc, showElement);
            createChild(ShowPOJO.FIELD_PERFORMANCES, show.getPerformances(), doc, showElement);
        }

        // Write DOM document to a file
        System.out.println(writeXml(doc));
    }

    // This method creates a child Element of the specific node name
    // and value, and appends it to the parent Element.
    private void createChild(String nodeName, Object value, Document doc, Element parent) {
        // If value is null, don't create an Element at all.
        if (value == null) {
            return;
        }
        String valueStr = "";

        // Determine the String representation of value.
        if (value instanceof String) {
            valueStr = (String) value;
        }
        if (value instanceof Date) {
            SimpleDateFormat fmt = new SimpleDateFormat("MM/dd/yyyy");
            valueStr = fmt.format((Date) value);
        }
        if (value instanceof Integer) {
            valueStr = value.toString();
        }

        // Create the Element, set its text value, and append it
        // to the parent.
        Element child = doc.createElement(nodeName);
        child.setTextContent(valueStr);
        parent.appendChild(child);
    }

    // This is an overload that creates nodes from a List of values.
    private void createChild(String parentName, String childName, List<String> valueList, Document doc, Element root) {
        Element parentElement = doc.createElement(parentName);
        root.appendChild(parentElement);
        for (String value : valueList) {
            createChild(childName, value, doc, parentElement);
        }
    }

    // Return XML from the supplied Document.
    private String writeXml(Document doc) throws TransformerException {

        TransformerFactory transformerFactory = TransformerFactory.newInstance();
        Transformer transformer = transformerFactory.newTransformer();

        // Set a property that "pretty prints" the output.
        transformer.setOutputProperty(OutputKeys.INDENT, "yes");

        DOMSource source = new DOMSource(doc);
        OutputStream output = new ByteArrayOutputStream();
        StreamResult result = new StreamResult(output);

        // Write the XML.
        transformer.transform(source, result);

        return output.toString();
    }
}
