package org.hardknockjava.lesson20.example01.reader.broadway;


/******************************************************************
 * This is the same as BroadwayReaderDOM, but instead of
 * taking in the entire input file at once, it breaks off
 * each <show> tag in turn and parses it.
 * On the output side, it writes the starting <broadway>
 * tag, then each <show> tag in turn, followed by the terminating
 * </broadway> tag.
 * 
 * Acknowledgments to Lokesh Gupta and Yong Mook Kim for code examples.
 */

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.Reader;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

public class BroadwayReaderDOMSingles {
    private static final String TAG_SHOW_START = "<show";
    private static final String TAG_SHOW_END = "</show>";
    
    // This statement gets the system line separator.
    private static final String LINE_SEPARATOR = System.getProperty("line.separator");

    private List<ShowPOJO> showList = new ArrayList<ShowPOJO>();

    private DocumentBuilderFactory factory = null;
    private DocumentBuilder builder = null;

    public static void main(String[] args) throws Exception {
        BroadwayReaderDOMSingles instance = new BroadwayReaderDOMSingles();

        // Read the file and store it in instances of ShowPOJO.
        instance.read("data/broadway.xml");

        // Dump the instances to the console.
        instance.dump();

        // Write the instances to the console.
        System.out.println();
        instance.write(System.out);
    }

    public BroadwayReaderDOMSingles() throws Exception {
        factory = DocumentBuilderFactory.newInstance();
        builder = factory.newDocumentBuilder();
    }

    // Read the XML and store it.
    private void read(String fileName) throws Exception {
        File showFile = new File(fileName);

        // Establish a Reader to read the file,
        // and a StringBuilder to hold its contents.
        Reader rdr = new FileReader(showFile);
        StringBuilder buf = new StringBuilder();

        // Read into the buffer until we hit a <show> tag.
        // Then delete all characters that preceded it and
        // read until we have a </show> tag. Present the
        // enclosed characters to the parser.
        String tag;
        while ((tag = extractShowTag(rdr, buf)) != null) {
            addToPojoList(tag);
        }
    }

    // Get the next <show> tag from the input.
    private String extractShowTag(Reader rdr, StringBuilder buf) throws IOException {
        String returnString = null;
        int start = -1;

        // Read more into the buffer until a start tag appears
        // or until the input is exhausted.
        char[] input = new char[60];
        while ((start = buf.indexOf(TAG_SHOW_START)) < 0) {
            int charsRead = rdr.read(input);
            if (charsRead < 0) {
                break;
            }
            buf.append(input, 0, charsRead);
        }

        // If no start tag is found, return a null.
        if (start < 0) {
            return null;
        }

        // Remove the characters up to the start tag.
        buf.delete(0, start);

        // Now search for the end tag.
        int end = -1;
        while ((end = buf.indexOf(TAG_SHOW_END)) < 0) {
            int charsRead = rdr.read(input);
            if (charsRead < 0) {
                break;
            }
            buf.append(input);
        }

        if (end < 0) {
            return null;
        }

        // Break off the complete <show> tag and return it.
        end += TAG_SHOW_END.length();

        returnString = buf.substring(0, end);
        buf.delete(0, end);
        return returnString;
    }

    // Parse the <show> tag and add the resulting ShowPOJO to
    // the list.
    private void addToPojoList(String tag) throws Exception {
        InputStream in = new ByteArrayInputStream(tag.getBytes());
        Document document = builder.parse(in);

        // nodeList contains all the <show> nodes. Of course,
        // there's only one.
        NodeList nodeList = document.getElementsByTagName("show");

        // Process the <show> node.
        Node node = nodeList.item(0);
        if (node.getNodeType() == Node.ELEMENT_NODE) {
            Element showElement = (Element) node;

            // The title is an Attribute of the <show> node.
            String title = showElement.getAttribute(ShowPOJO.FIELD_TITLE);

            // The remaining fields are elements nested within the <show> node.
            List<String> playwrightList = getListValue(ShowPOJO.FIELD_PLAYWRIGHTS, ShowPOJO.FIELD_PLAYWRIGHT,
                    showElement);
            String theatre = getNodeStringValue(ShowPOJO.FIELD_THEATRE, showElement);
            List<String> composerList = getListValue(ShowPOJO.FIELD_COMPOSERS, ShowPOJO.FIELD_COMPOSER,
                    showElement);
            List<String> lyricistList = getListValue(ShowPOJO.FIELD_LYRICISTS, ShowPOJO.FIELD_LYRICIST,
                    showElement);
            Date opening = getNodeDateValue(ShowPOJO.FIELD_OPENING, showElement);
            Date closing = getNodeDateValue(ShowPOJO.FIELD_CLOSING, showElement);
            Integer previews = getNodeIntegerValue(ShowPOJO.FIELD_PREVIEWS, showElement);
            Integer performances = getNodeIntegerValue(ShowPOJO.FIELD_PERFORMANCES, showElement);

            ShowPOJO show = new ShowPOJO(title, theatre, playwrightList, composerList, lyricistList, opening,
                    closing, previews, performances);
            showList.add(show);
        }
    }

    // This method returns the text contained within an Element.
    // Note that there's no way to get a single element by name.
    // The DOM assumes there might be more than one. But we're
    // interested in only the first even if there are more.
    private String getNodeStringValue(String elementName, Element element) {
        String returnValue = null;
        NodeList list = element.getElementsByTagName(elementName);
        if (list.getLength() > 0) {
            Node node = list.item(0);
            returnValue = node.getTextContent();
        }

        return returnValue;
    }

    private List<String> getListValue(String parentName, String childName, Element element) {
        List<String> returnValue = new ArrayList<String>();

        NodeList parentList = element.getElementsByTagName(parentName);

        for (int parentIndex = 0; parentIndex < parentList.getLength(); parentIndex++) {
            Node parentNode = parentList.item(parentIndex);
            NodeList childList = parentNode.getChildNodes();
            for (int childIndex = 0; childIndex < childList.getLength(); childIndex++) {
                Node childNode = childList.item(childIndex);
                if (childNode.getChildNodes().getLength() > 0) {
                    returnValue.add(childNode.getTextContent());
                }
            }
        }

        return returnValue;
    }

    // This method is for elements that contain the string representation
    // of dates.
    private Date getNodeDateValue(String elementName, Element element) {
        String dateStr = getNodeStringValue(elementName, element);
        Date returnValue = null;

        if (dateStr != null) {
            SimpleDateFormat fmt = new SimpleDateFormat("MM/dd/yyyy");
            try {
                returnValue = fmt.parse(dateStr);
            } catch (Exception e) {
                returnValue = null;
            }
        }

        return returnValue;
    }

    // This method is for elements that contain the string representation
    // of Integers.
    private Integer getNodeIntegerValue(String elementName, Element element) {
        String integerStr = getNodeStringValue(elementName, element);
        Integer returnValue = null;

        if (integerStr != null) {
            try {
                returnValue = Integer.valueOf(integerStr);
            } catch (Exception e) {
                returnValue = null;
            }
        }

        return returnValue;
    }

    // Dump the list of Shows to the console.
    private void dump() {
        for (ShowPOJO show : showList) {
            System.out.println(show.toString());
        }
    }

    // Format the list of ShowPOJOs in XML, surrounded by
    // the outer <broadway> tags, and write to the console
    // one Production at a time.
    private void write(OutputStream out) throws Exception {
        // Since we're not having the DOM write the outer tags,
        // we need to write the XML declaration here. This code
        // will accomplish that.
        {
            Document doc = builder.newDocument();
            
            // Write the document to the output stream with no tags.
            // Only the XML declaration will be written.
            out.write(formatXml(doc, false).getBytes());
        }
        
        out.write(LINE_SEPARATOR.getBytes());
        out.write("<broadway>".getBytes());
        out.write(LINE_SEPARATOR.getBytes());

        // root elements
        for (int index = showList.size() - 1; index >= 0; index--) {
            Document doc = builder.newDocument();

            ShowPOJO show = showList.get(index);
            Element rootElement = doc.createElement("show");
            rootElement.setAttribute(ShowPOJO.FIELD_TITLE, show.getTitle());
            doc.appendChild(rootElement);

            createChild(ShowPOJO.FIELD_THEATRE, show.getTheatre(), doc, rootElement);
            createChild(ShowPOJO.FIELD_PLAYWRIGHTS, ShowPOJO.FIELD_PLAYWRIGHT, show.getPlaywrightList(), doc,
                    rootElement);
            createChild(ShowPOJO.FIELD_COMPOSERS, ShowPOJO.FIELD_COMPOSER, show.getComposerList(), doc,
                    rootElement);
            createChild(ShowPOJO.FIELD_LYRICISTS, ShowPOJO.FIELD_LYRICIST, show.getLyricistList(), doc,
                    rootElement);
            createChild(ShowPOJO.FIELD_OPENING, show.getOpening(), doc, rootElement);
            createChild(ShowPOJO.FIELD_CLOSING, show.getClosing(), doc, rootElement);
            createChild(ShowPOJO.FIELD_PREVIEWS, show.getPreviews(), doc, rootElement);
            createChild(ShowPOJO.FIELD_PERFORMANCES, show.getPerformances(), doc, rootElement);
            out.write(formatXml(doc, true).getBytes());
        }

        // Write DOM document to a file
        out.write("</broadway>".getBytes());
    }

    // This method creates a child Element of the specific node name
    // and value, and appends it to the parent Element.
    private void createChild(String nodeName, Object value, Document doc, Element parent) {
        // If value is null, don't create an Element at all.
        if (value == null) {
            return;
        }
        String valueStr = "";

        // Determine the String representation of value.
        if (value instanceof String) {
            valueStr = (String) value;
        }
        if (value instanceof Date) {
            SimpleDateFormat fmt = new SimpleDateFormat("MM/dd/yyyy");
            valueStr = fmt.format((Date) value);
        }
        if (value instanceof Integer) {
            valueStr = value.toString();
        }

        // Create the Element, set its text value, and append it
        // to the parent.
        Element child = doc.createElement(nodeName);
        child.setTextContent(valueStr);
        parent.appendChild(child);
    }

    // This is an overload that creates nodes from a List of values.
    private void createChild(String parentName, String childName, List<String> valueList, Document doc, Element root) {
        if (valueList.isEmpty()) {
            return;
        }

        Element parentElement = doc.createElement(parentName);
        root.appendChild(parentElement);
        for (String value : valueList) {
            createChild(childName, value, doc, parentElement);
        }
    }

    // Return XML from the supplied Document.
    private String formatXml(Document doc, boolean omitXmlDeclaration)
            throws TransformerException {

        TransformerFactory transformerFactory = TransformerFactory.newInstance();
        Transformer transformer = transformerFactory.newTransformer();

        // This method creates an internal tag, so we don't want the XML
        // declaration tag to appear here. We'd also like "pretty print."
        if (omitXmlDeclaration) {
            transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "yes");

            // Set a property that "pretty prints" the output.
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");
        }

        DOMSource source = new DOMSource(doc);
        OutputStream output = new ByteArrayOutputStream();
        StreamResult result = new StreamResult(output);

        // Write the XML.
        transformer.transform(source, result);
        
        return output.toString();
    }
}
